/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer & David Odin
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <quaternion.hpp>
#include <matrix3.hpp>
#include <vec3.hpp>
#include <cmath>

namespace cpe
{
quaternion::quaternion()
    :vec4(1,0,0,0)
{}
quaternion::quaternion(float x_param,float y_param,float z_param,float w_param)
    :vec4(x_param,y_param,z_param,w_param)
{}

quaternion::quaternion(const vec3& axis_param,float angle_param)
{
    float sin_phi=std::sin(angle_param/2.0);
    float cos_phi=std::cos(angle_param/2.0);
    vec3 n_axis=axis_param.normalized();

    x() = n_axis[0]*sin_phi;
    y() = n_axis[1]*sin_phi;
    z() = n_axis[2]*sin_phi;
    w() =           cos_phi;
}

quaternion::quaternion(const vec4& v)
    :vec4(v)
{}

vec3 quaternion::axis() const
{
    return vec3(x(),y(),z()).normalized();
}
float quaternion::angle() const
{
    return 2*std::acos(w());
}

matrix3 quaternion::matrix() const
{
    float x2=x()*x();
    float y2=y()*y();
    float z2=z()*z();
    float xy=x()*y();
    float xz=x()*z();
    float yz=y()*z();
    float wx=w()*x();
    float wy=w()*y();
    float wz=w()*z();

    return matrix3(1-2*(y2+z2),  2*(xy-wz),  2*(xz+wy),
                   2*(xy+wz),1-2*(x2+z2),  2*(yz-wx),
                   2*(xz-wy),  2*(yz+wx),1-2*(x2+y2));
}

quaternion quaternion::conjugated() const
{
    return quaternion(vec4(-x(),-y(),-z(),w()));
}

quaternion quaternion::slerp(const quaternion& q0,const quaternion& q1,const float t)
{
#if 0
    quaternion q0_local=q0;
    float epsilon=1e-6;

    //do not interpolate for very small differences
    if( (q0-q1).norm()<epsilon )
        return q0;

    float cos_theta = q0.dot(q1);
    if(cos_theta<0)
    {
        q0_local=-q0;
        cos_theta*=-1;
    }



    float sin_theta = sqrt(1-cos_theta*cos_theta);
    float theta = atan2(sin_theta,cos_theta);

    //linear interpolation for small angles
    if(std::abs(sin_theta)<epsilon)
        return (1-t)*q0_local+t*q1;

    const float w0=sin((1-t) *theta)/sin_theta;
    const float w1=sin(  t   *theta)/sin_theta;

    quaternion interpolated = w0*q0_local+w1*q1;

    return interpolated;
#endif
    if (t <= 0)
      return q0;
    if (t>=1)
      return q1;
    float cos_omega = q0.dot(q1);
    float q1w = q1.w();
    float q1x = q1.x();
    float q1y = q1.y();
    float q1z = q1.z();
    if (cos_omega < 0)
    {
      q1w = -q1w;
      q1x = -q1x;
      q1y = -q1y;
      q1z = -q1z;
      cos_omega = -cos_omega;
    }
    float k0, k1;
    if (cos_omega > 0.9999f)
    {
      k0 = 1.0f - t; k1 = t;
    }
    else
    {
      float sin_omega = sqrt(1.0f - cos_omega * cos_omega);
      float omega = atan2(sin_omega, cos_omega);
      float one_over_sin_omega = 1.0f / sin_omega;
      k0 = sin((1.0f - t) * omega) * one_over_sin_omega;
      k1 = sin(t * omega) * one_over_sin_omega;
    }
    quaternion out;
    out.w() = (k0 * q0.w()) + (k1 * q1w);
    out.x() = (k0 * q0.x()) + (k1 * q1x);
    out.y() = (k0 * q0.y()) + (k1 * q1y);
    out.z() = (k0 * q0.z()) + (k1 * q1z);
    return out;
}

vec3 quaternion::rotate(const vec3& v) const
{
    float x2=x()*x();
    float y2=y()*y();
    float z2=z()*z();
    float xy=x()*y();
    float xz=x()*z();
    float yz=y()*z();
    float wx=w()*x();
    float wy=w()*y();
    float wz=w()*z();

    //identical to (*this)*vec*quaternion(-x,-y,-z,w);
    return vec3( (1-2*(y2+z2))*v.x() + 2*(xy-wz)*v.y() + 2*(xz+wy)*v.z(),
               2*(xy+wz)*v.x()+(1-2*(x2+z2))*v.y()+2*(yz-wx)*v.z(),
               2*(xz-wy)*v.x()+2*(yz+wx)*v.y()+(1-2*(x2+y2))*v.z() );
}

quaternion& quaternion::operator*=(const quaternion& q)
{
    //identical to (w0*v1+w1*v0-v0xv1,w0*w1-v0.v1), with v0=(x,y,z) & v1=(q.x,q.y,q.z).

    *this =quaternion(  (x() * q.w()) + (w() * q.x()) + (y() * q.z()) - (z() * q.y()),
                        (y() * q.w()) + (w() * q.y()) + (z() * q.x()) - (x() * q.z()),
                        (z() * q.w()) + (w() * q.z()) + (x() * q.y()) - (y() * q.x()),
                        (w() * q.w()) - (x() * q.x()) - (y() * q.y()) - (z() * q.z()) );


    return *this;
}
quaternion& quaternion::operator*=(float s)
{
    x()*=s;
    y()*=s;
    z()*=s;
    w()*=s;
    return *this;
}

quaternion& quaternion::operator+=(const quaternion& q)
{
    x()+=q.x();
    y()+=q.y();
    z()+=q.z();
    w()+=q.w();
    return *this;
}
quaternion& quaternion::operator-=(const quaternion& q)
{
    x()-=q.x();
    y()-=q.y();
    z()-=q.z();
    w()-=q.w();
    return *this;
}
quaternion& quaternion::operator/=(float s)
{
    float epsilon=1e-8;
    if(std::abs(s)<epsilon)
        throw cpe::exception_vec4("Division par 0",__FILE__,__FUNCTION__,__LINE__);

    x()/=s;
    y()/=s;
    z()/=s;
    w()/=s;
    return *this;
}

quaternion quaternion::operator*(const quaternion& q) const
{
    quaternion temp=*this;
    temp*=q;
    return temp;
}
quaternion quaternion::operator*(const vec3& v) const
{
    return quaternion(  (w() * v.x()) + (y() * v.z()) - (z() * v.y()),
                        (w() * v.y()) + (z() * v.x()) - (x() * v.z()),
                        (w() * v.z()) + (x() * v.y()) - (y() * v.x()),
                       -(x() * v.x()) - (y() * v.y()) - (z() * v.z()) );
}
quaternion quaternion::operator*(float s) const
{
    quaternion temp=*this;
    temp*=s;
    return temp;
}
quaternion operator*(float s,const quaternion& q)
{
    return q*s;
}
quaternion quaternion::operator+(const quaternion& q) const
{
    quaternion temp=*this;
    temp+=q;
    return temp;
}
quaternion quaternion::operator-(const quaternion& q) const
{
    quaternion temp=*this;
    temp-=q;
    return temp;
}
quaternion quaternion::operator/(float s) const
{
    quaternion temp=*this;
    temp/=s;
    return temp;
}

quaternion quaternion::operator-() const
{
    return quaternion(vec4(-x(),-y(),-z(),-w()));
}


}
