/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef VEC2_HPP
#define VEC2_HPP

#include <string>
#include <iostream>

#include <exception_cpe.hpp>

namespace cpe
{

/** \brief Vectors/Points 2D */
class vec2
{
public:

    // ********************************************* //
    // ********************************************* //
    //  CONSTRUCTORS
    // ********************************************* //
    // ********************************************* //

    /** \brief empty constructor */
    vec2();
    /** \brief direct constructor */
    vec2(float x,float y);

    // ********************************************* //
    // ********************************************* //
    //  Get/set
    // ********************************************* //
    // ********************************************* //

    /** \brief get x coordinate */
    float x() const;
    /** \brief get x coordinate */
    float& x();
    /** \brief get y coordinate */
    float y() const;
    /** \brief get y coordinate */
    float& y();


    /** \brief Access to the k_th entry (k in [0,1]) */
    float operator[](const size_t& k) const;
    /** \brief Access to the k_th entry (k in [0,1]) */
    float& operator[](const size_t& k);
    /** \brief Access to the k_th entry (k in [0,1]) */
    float operator()(const size_t& k) const;
    /** \brief Access to the k_th entry (k in [0,1]) */
    float& operator()(const size_t& k);

    /** \brief set every entry to 0 */
    void set_zero();


    /** \brief fast pointer access
        \warning unprotected access */
    const float *pointer() const;

    // ********************************************* //
    // ********************************************* //
    //  Math operation
    // ********************************************* //
    // ********************************************* //

    /** \brief perform dot product between two v2 */
    float dot(const vec2& p) const;

    /** \brief get the norm of the vector */
    float norm() const;
    /** \brief get the square norm of the vector */
    float norm2() const;
    /** \brief normalize the vector to unit length */
    vec2 normalized() const;




    // ********************************************* //
    // ********************************************* //
    //  Operator +-*/
    // ********************************************* //
    // ********************************************* //

    /** \brief + operator */
    vec2 operator+(const vec2& p2) const;
    /** \brief + operator */
    vec2 operator+(float s) const;

    /** \brief - operator */
    vec2 operator-(const vec2& p2) const;
    /** \brief - operator */
    vec2 operator-(float s) const;


    /** \brief multiply by a scalar operator */
    vec2 operator*(float s) const;
    /** \brief divide by a scalar operator */
    vec2 operator/(float s) const;

    /** \brief internal + */
    vec2& operator+=(const vec2& p);
    /** \brief internal + */
    vec2& operator+=(float s);
    /** \brief internal - */
    vec2& operator-=(const vec2& p);
    /** \brief internal - */
    vec2& operator-=(float s);
    /** \brief internal * */
    vec2& operator*=(float s);
    /** \brief internal / */
    vec2& operator/=(float s);

    /** \brief unary negation */
    vec2 operator-() const;

    /** \brief does componentwise mutliplication
            \return MC_v2d(x*p.x(),y*p.y(),z*p.z())
        */
    vec2 product_compontentwise(const vec2& p) const;
    /** \brief does componentwise mutliplication

            perform p=(p.x()*p2.x(),p.y()*p2.x())
        */
    vec2& product_compontentwise_internal(const vec2& p);
    /** \brief internal scaling (similar to componentwise) */
    void scale(float sx,float sy);


    // ********************************************* //
    // ********************************************* //
    //  Output
    // ********************************************* //
    // ********************************************* //

    /** \brief export the value as string cout<<v2(2,3,6) => 2 3 6 */
    std::string to_string() const;



private:


    // ********************************************* //
    // ********************************************* //
    //  Internal parameters
    // ********************************************* //
    // ********************************************* //

    /** \brief x coordinate */
    float internal_x;
    /** \brief y coordinate */
    float internal_y;


    // ********************************************* //
    // ********************************************* //
    //  Helper
    // ********************************************* //
    // ********************************************* //

    /** \brief assert that a size_t belongs to [[0,1]] */
    void assert_size(const size_t& k) const;

};

/** \brief + operator */
vec2 operator+(float s,const vec2& p);
/** \brief - operator */
vec2 operator-(float s,const vec2& p);
/** \brief multiply by a scalar operator */
vec2 operator*(float s,const vec2& p);

// ********************************************* //
// ********************************************* //
//  Output
// ********************************************* //
// ********************************************* //

/** \brief output the vector in ostream as (x,y,z) */
std::ostream& operator<<(std::ostream& stream,const vec2& p);


/** \brief Special exception class for v2 */
class exception_vec2 : public exception_cpe
{
public:

    /** \brief empty constructor */
    exception_vec2():exception_cpe(){}
    /** \brief constructor with error information */
    exception_vec2(const std::string& msg_param,const std::string& file_param,const std::string& caller_param,const int& line_param):exception_cpe(msg_param,file_param,caller_param,line_param){}
};

}



#endif
