/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef IMAGE_HPP
#define IMAGE_HPP

#include <iostream>
#include <vector>
#include <exception_cpe.hpp>
#include <color.hpp>



namespace cpe
{

class pos2;

/** \brief Container for picture data. Data are stored as unsigned char [0,255] */
class image
{
    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief empty constructor */
        image();
        /** \brief create square picture of size NxN
            \param unsigned int N: size (width & height) of the picture
        */
        image(int N);
        /** \brief create picture of size Nx x Ny
            \param unsigned int Nx: picture width
            \param unsigned int Ny: picture height
        */
        image(int Nx,int Ny);
        /** \brief import pre-existing picture from file filename
            \param std::string filename: path to the file to load
        */
        image(const std::string& filename);


        // ********************************************* //
        // ********************************************* //
        //  Picture Size
        // ********************************************* //
        // ********************************************* //

        /** \brief get the picture width in pixel */
        int Nx() const;
        /** \brief get the picture height in pixel */
        int Ny() const;

        /** \brief try to reshape the picture to a new size
            \param unsigned int Nx: the new picture width
            \note Nx x Ny / Nx2 must be integer
        */
        void reshape(int Nx2);


        /** \brief check if a given pixel (kx,ky) is within the size of the picture
            \param p2d u: current (kx,ky)-coordinate
            \return true if (kx,ky) is within the picture, false otherwise
        */
        bool check_position(const pos2& u) const;


        // ********************************************* //
        // ********************************************* //
        //  Pixel access
        // ********************************************* //
        // ********************************************* //

        /** \brief Accessor on a pixel */
        const color& operator()(const pos2& index) const;
        /** \brief Accessor on a pixel */
        color& operator()(const pos2& index);


        // ********************************************* //
        // ********************************************* //
        //  Special pixel filling
        // ********************************************* //
        // ********************************************* //

        /** \brief fill the entire picture with a given (r,g,b) color
            \param color c: (red,green,blue) value
        */
        void fill(const color& c);
        /** \brief fill a rectangle ((x1,y1);(x2,y2)) with a given (r,g,b) color
            \param p2d u1: (x,y)-coordinate of starting point
            \param p2d u2: (x,y)-coordinate of ending point
            \param color c: (r,g,b) value
        */
        void fill_rectangle(const pos2& u1,const pos2& u2,const color& c);


        // ********************************************* //
        // ********************************************* //
        //  I/O
        // ********************************************* //
        // ********************************************* //

        /** \brief write the picture into a file */
        void export_file(const std::string& filename);



    protected:

        /** \brief Ensure a position is within the size of the image, otherwise throw exception
        */
        void assert_position(const pos2& u) const;


        /** \brief transform coordinate u=(kx,ky) to index in vector data */
        int coordinate_to_index(const pos2& u) const;

    private:


        // ********************************************* //
        // ********************************************* //
        //  Internal data
        // ********************************************* //
        // ********************************************* //

        /** \brief internal data storage as std::vector of unsigned char

            Picture is stored as (color_0 , color_1 , ...) vector size is therefore 3 Nx Ny
        */
        std::vector<color> internal_data;
        /** \brief storage of picture width */
        int internal_Nx;
        /** \brief storage of picture height */
        int internal_Ny;


};


/** \brief Special exception class for image */
class exception_image : public exception_cpe
{
    public:

        /** \brief empty constructor */
        exception_image():exception_cpe(){}
        /** \brief constructor with error information */
        exception_image(const std::string& msg_param,const std::string& file_param,const std::string& caller_param,const int& line_param):exception_cpe(msg_param,file_param,caller_param,line_param){}
};
}

#endif
