/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <triangle_scanline.hpp>
#include <assert.h>


namespace cpe
{

line_pointer::line_pointer(int line_number_data,int index_data)
    :line_number(line_number_data),index(index_data)
{}

const pos2& triangle_scanline::get_coordinate(const line_pointer& p)
{
    //protection against index outside vector dimension
    assert(p.line_number>=0 && p.line_number<3);
    assert(line[p.line_number].size()>p.index && p.index>=0);

    //get coordinate
    return line[p.line_number].coordinate[p.index];
}

int triangle_scanline::size()
{
    return first_point.size();
}

triangle_scanline::triangle_scanline(const pos2& u0,const pos2& u1,const pos2& u2)
{
    // ********************************************* //
    // ********************************************* //
    // Algorithm:
    //
    // Compute the 3 lines [u0,u1] , [u1,u2] , [u2,u0]
    // Compute interpolation values for the 3 lines
    //
    // (x_min,x_max)=min/max of x values of the 3 lines
    //
    // Init pointers table to any values corresponding to a given x-coordinate
    //
    // for all 3 lines
    //     for all index in line
    //
    //         (x,y_current) := current value of the line
    //         y_first := current y value pointed by x in the first_point table
    //         y_last  := current y value pointed by x in the last_point table
    //
    //         if(y_first>y_current) update first_point table to y_current
    //         if(y_last <y_current) update last_point table to y_current
    //
    // ********************************************* //
    // ********************************************* //



    //Compute the 3 discrete lines
    line[0]=compute_line(u0,u1);
    line[1]=compute_line(u1,u2);
    line[2]=compute_line(u2,u0);

    //Compute interpolation for the 3 lines
    for(int k_line=0;k_line<3;++k_line)
        line[k_line].compute_interpolation();

    //Get x_min & x_max
    int x_min=std::min(std::min(u0.x(),u1.x()),u2.x());
    int x_max=std::max(std::max(u0.x(),u1.x()),u2.x());

    //init vector size
    int delta_x=x_max-x_min+1;
    first_point.resize(delta_x);
    last_point.resize(delta_x);

    // Init pointers table to any values corresponding to a given x-coordinate
    for(int k_line=0;k_line<3;++k_line)
    {
        for(int k=0;k<line[k_line].size();++k)
        {
            const pos2& p=line[k_line].coordinate[k];
            int offset=p.x()-x_min;
            first_point[offset]=line_pointer(k_line,k);
            last_point[offset]=line_pointer(k_line,k);
        }
    }

    // for all 3 lines
    //     for all index in line
    //
    //         (x,y_current) := current value of the line
    //         y_first := current y value pointed by x in the first_point table
    //         y_last  := current y value pointed by x in the last_point table
    //
    //         if(y_first>y_current) update first_point table to y_current
    //         if(y_last <y_current) update last_point table to y_current
    for(int k_line=0;k_line<3;++k_line)
    {
        for(int k=0;k<line[k_line].size();++k)
        {
            const pos2& p=line[k_line].coordinate[k];
            int offset=p.x()-x_min; //current index in the vector
            int y=p.y(); //current y

            //first point
            const pos2& current_coordinate_first=get_coordinate(first_point[offset]);
            if(current_coordinate_first.y()>=y) //update first_point table
                first_point[offset]=line_pointer(k_line,k);

            //last point
            const pos2& current_coordinate_last=get_coordinate(last_point[offset]);
            if(current_coordinate_last.y()<=y) //update last_point table
                last_point[offset]=line_pointer(k_line,k);

        }
    }

}

}
